( cd usr/lib64 ; rm -rf libdecoration.so.0 )
( cd usr/lib64 ; ln -sf libdecoration.so.0.0.0 libdecoration.so.0 )
( cd usr/lib64 ; rm -rf libdecoration.so )
( cd usr/lib64 ; ln -sf libdecoration.so.0.0.0 libdecoration.so )
